# coding=utf-8
#
# created by kpe on 09.01.2021 at 4:47 PM
#

from __future__ import division, absolute_import, print_function

import os
import re
import time
import subprocess
import logging

from aiogram import Bot, Dispatcher, executor, types, filters
import pyotp
import params as pp


log = logging.getLogger(__name__)

REX_AUTH = r'/auth\s+([0-9]{6})'
REX_DO = r'/do\s+([a-z0-9]+)\s?(.*)'


class TeleRemBot(pp.WithParams):
    class Params(pp.Params):
        bot_name = pp.Param("TeleRemBash", dtype=str, doc="The Bot Name (used in the OTP provisioning uri only)")
        api_token = pp.Param(None, dtype=str, doc="Telegram Bot API Token")
        auth_secret = pp.Param(None, dtype=str, doc="OPT Authenticator Secret (RFC 6238)")
        scripts_root = pp.Param("scripts", dtype=str, doc="scripts location")
        scripts_timeout = pp.Param(30, dtype=int, doc="execution timeout in seconds")
        username = pp.Param(None, dtype=str, doc="Username of the authorized user")
        user_id = pp.Param(None, dtype=int, doc="User IDs of the authorized user")

        @staticmethod
        def resolve_path(path):
            return path if path.startswith('/') else os.path.join(os.getcwd(), path)

        def get_scripts_root(self):
            return self.resolve_path(str(self.scripts_root))

    @property
    def params(self) -> Params:
        return self._params

    def _construct(self, *args, **kwargs):
        super(TeleRemBot, self)._construct(*args, **kwargs)
        if self.params.auth_secret is None:
            raise AttributeError("AUTH_SECRET not specified")
        self.totp = pyotp.TOTP(self.params.auth_secret)

        self.user_whitelist = [self.params.username]
        self.user_id_whitelist = [self.params.user_id]
        self.chat_id_whitelist = []

        # Initialize bot and dispatcher
        self.bot = Bot(token=self.params.api_token)
        self.dp = Dispatcher(self.bot)

        self.dp.register_message_handler(self.cmd_auth,
                                         filters.RegexpCommandsFilter(regexp_commands=[REX_AUTH]))
        self.dp.register_message_handler(self.cmd_do_execute,
                                         filters.RegexpCommandsFilter(regexp_commands=[REX_DO]))
        self.dp.register_message_handler(self.default_message_handler)

    def _precondition_fail(self, message: types.Message, field: str, value: str):
        log.warning(f"pre-condition fail: {field}:[{value}]: {message.to_python()}")

    def check_preconditions(self, message: types.Message, allow_unauthorized=False) -> bool:
        user_id = message.from_user.id
        user_name = message.from_user.username
        chat_id = message.chat.id

        if user_name not in self.user_whitelist:
            self._precondition_fail(message, 'user_name', user_name)
            return False
        if self.user_id_whitelist and user_id not in self.user_id_whitelist:
            self._precondition_fail(message, 'user_id', user_id)
            return False
        if not allow_unauthorized and chat_id not in self.chat_id_whitelist:
            self._precondition_fail(message, 'chat_id', user_id)
            return False
        if abs(int(message.date.timestamp()) - time.time()) > 10:
            self._precondition_fail(message, 'date', str(message.date.timestamp()))
            return False

        return True

    # @dp.message_handler(regex '/do\s+([0-9]+)')
    async def cmd_auth(self, message: types.Message, regexp_command: re.Match):
        if not self.check_preconditions(message, allow_unauthorized=True):
            return

        token = regexp_command.group(1)
        if not self.totp.verify(token):
            await message.answer("invalid")
            return

        username = message.from_user.username
        chat_id = message.chat.id
        for cid in self.chat_id_whitelist:   # notify current chats
            if cid != chat_id:
                await self.bot.send_message(chat_id, f"Farewell! Serving {username}!")
        self.chat_id_whitelist = [chat_id]

        await message.answer(f"Welcome, {username}!")
        await self.execute_script(message, 'welcome', '', silent_if_not_found=True)

    # @dp.message_handler(filters.RegexpCommandsFilter(regexp_commands=[r'/do\s+([a-z0-9]+)\s?(.*)']))
    async def cmd_do_execute(self, message: types.Message, regexp_command: re.Match):
        if not self.check_preconditions(message):
            return

        cmd = regexp_command.group(1)
        params = regexp_command.group(2)
        await self.execute_script(message, cmd, params)

    async def execute_script(self, message: types.Message, cmd: str, params: str, silent_if_not_found=False):
        script_root = self.params.get_scripts_root()
        script_file = os.path.join(script_root, f"{cmd}")

        # auto append .sh suffix if there is no ext in the command
        if not os.path.exists(script_file) and len(cmd.split(".")) == 1 and os.path.exists(script_file + ".sh"):
            script_file += ".sh"
        else:
            if not silent_if_not_found:
                await message.reply(f"Can't do [{cmd}]")
            return

        try:
            out = subprocess.check_output([script_file] + params.split(),
                                          timeout=self.params.scripts_timeout,
                                          cwd=self.params.resolve_path(self.params.scripts_root),
                                          )
            await message.answer(out.decode('utf8'))
        except Exception as ex:
            log.warning(f"FAILED executing {script_file} {params.split()}: {ex}")
            await message.reply("failed")

    # @dp.message_handler()
    async def default_message_handler(self, message: types.Message):
        if not self.check_preconditions(message):
            return
        await message.reply("Don't understand")

    def main(self):
        import telerembash as tb
        print(f"TeleRemBash v{tb.__version__}")
        executor.start_polling(self.dp, skip_updates=True)



# coding=utf-8
#
# created by kpe on 09.01.2021 at 2:22 PM
#

from __future__ import division, absolute_import, print_function

from .version import __version__
from ._logging import setup_logging

import os
setup_logging(default_path=os.path.normpath(os.path.join(os.getcwd(), "logging.config.yaml")))



# coding=utf-8
#
# created by kpe on 10.01.2021 at 12:55 PM
#

from __future__ import division, absolute_import, print_function

import os
import sys
import logging

import params as pp
import daemon
import lockfile

from telerembash.bot import TeleRemBot


log = logging.getLogger(__name__)


class BotDaemonCli(pp.WithParams):
    class Params(pp.Params):
        daemon = pp.Param(None, dtype=str, doc=f"first positional should be daemon", positional=True)
        command = pp.Param(None, dtype=str, doc=f"one of - start, stop", positional=True)
        config = pp.Param("/etc/teleremd.config.yaml", dtype=str, doc="teleremd daemon config")

    @property
    def params(self) -> Params:
        return self._params

    def main(self):
        if not os.path.isabs(self.params.config) or not os.path.exists(self.params.config):
            print(f"invalid config path: {self.params.config}")
            exit(3)
        params = BotDaemon.Params.from_yaml_file(self.params.config)
        daemon_cli = BotDaemon.from_params(params)
        cli_cmd = self.params.command
        if cli_cmd in ["start"]:
            daemon_cli.start()
        elif cli_cmd in ["stop"]:
            daemon_cli.stop()
        else:
            print(f"unknown daemon command: {cli_cmd}")
            exit(4)


class BotDaemon(pp.WithParams):
    class Params(TeleRemBot.Params):
        service_name = pp.Param("teleremd", dtype=str, doc=f"(install only) name of the service/daemon")
        service_user = pp.Param("teleremd", dtype=str, doc=f"(install only) user for the service/daemon")
        pid_file = pp.Param(None, dtype=str, doc=f"pidfile for the daemon")

        def get_home_dir(self):
            import pwd
            home_dir = pwd.getpwnam(self.service_user).pw_dir
            return home_dir

        def get_pid_file(self):
            pid_file = f"/run/{self.service_name}/{self.service_name}.pid" if self.pid_file is None else self.pid_file
            return pid_file

    def start(self):
        import pwd
        params = self.params
        with daemon.DaemonContext(
            # pidfile=lockfile.FileLock(params.get_pid_file()),
            working_directory=params.get_scripts_root(),
            uid=pwd.getpwnam(params.service_user).pw_uid,
            umask=0o0133,
            detach_process=True,

            stderr=sys.stderr,
            stdout=sys.stdout
        ):
            with open(params.get_pid_file(), "wt") as pid_file:
                print(str(os.getpid()), file=pid_file)

            bot = TeleRemBot.from_params(params)
            bot_params = dict(filter(lambda x: x[0] not in ['api_token', 'auth_secret'], bot.params.items()))
            log.info(f"starting daemon bot {bot_params}")
            bot.main()


def main():
    pass


if __name__ == "__main__":
    main()



# coding=utf-8
#
# created by kpe on 16.10.2020 at 10:04 PM
#

from __future__ import division, absolute_import, print_function

import os
import sys
import yaml
import logging
import logging.config
import coloredlogs


def setup_logging(default_path='logging.yaml', default_level=logging.INFO, env_key='LOG_CFG'):

    path = default_path
    value = os.getenv(env_key, None)
    if value:
        path = value
    if os.path.exists(path):
        with open(path, 'rt') as f:
            try:
                config = yaml.safe_load(f.read())
                logging.config.dictConfig(config)
            except Exception as e:
                print(e)
                print('Error in Logging Configuration. Using default configs')
                logging.basicConfig(level=default_level)
                coloredlogs.install(level=default_level)
    else:
        logging.basicConfig(level=default_level)
        coloredlogs.install(level=default_level)
        print(f"Failed to load logging configuration:[{path}]. Using default configs", file=sys.stderr)



# coding=utf-8
#
# created by kpe on 10.01.2021 at 12:55 PM
#

from __future__ import division, absolute_import, print_function

import os
import re
import sys
import json
import string
import logging
import subprocess
from collections import defaultdict

import params as pp
import pyotp

from telerembash.bot import TeleRemBot
from telerembash.daemon import BotDaemon


log = logging.getLogger(__name__)

COMMANDS = ['init', 'start', 'install']


class BotCli(pp.WithParams):
    class Params(TeleRemBot.Params):
        command = pp.Param(None, dtype=str, doc=f"command - one of: {COMMANDS}", positional=True)
        config  = pp.Param("telerem.config.yaml", dtype=str, doc="config yaml file location")
        systemd = pp.Param(False, dtype=bool, doc=f"installs a systemd service")
        initd   = pp.Param(False, dtype=bool, doc=f"installs an init.d daemon script")

    @property
    def params(self) -> Params:
        return self._params

    def _construct(self):
        self.command_handlers = defaultdict(lambda: self._cmd_not_found)
        self.command_handlers['init'] = self._do_init
        self.command_handlers['start'] = self._do_start
        self.command_handlers['install'] = self._do_install

    def _cmd_not_found(self):
        log.error(f"Unrecognized command:[{self.params.command}]")

    def main(self):
        # pretty print non null arguments
        log.debug("main({})".format(json.dumps(dict(filter(lambda t: t[1] is not None,
                                                           dict(self.params).items())),
                                               indent=2)))
        # execute cli command
        self.command_handlers[self.params.command]()

    def _load_config(self):
        def filter_empty_values(d: dict):
            return dict(filter(lambda t: t[1] is not None, d.items()))

        config = TeleRemBot.Params()
        config_path = self.params.config
        config_path = self.params.resolve_path(config_path)
        if os.path.isfile(config_path):
            log.info(f"Loading existing config at:[{config_path}]")
            config = TeleRemBot.Params.from_yaml_file(config_path)
        else:
            log.info(f"Creating new config at:[{config_path}]")
        config = filter_empty_values(dict(config))
        config.update(filter_empty_values(dict(self.params)))
        config = TeleRemBot.Params.from_dict(config, return_instance=True, return_unused=False)
        return config

    def _do_init(self):
        config = self._load_config()
        if config.auth_secret is None:
            log.warning("generating new AUTH_SECRET")
            config.auth_secret = pyotp.random_base32()
            self._provision_totp_secret(config.auth_secret)

        if config.api_token is None:
            log.warning("API_TOKEN was not specified")

        config_path = self.params.resolve_path(self.params.config)
        config.to_yaml_file(config_path)
        log.debug(f"writing config to: {config_path}")

    def _provision_totp_secret(self, auth_secret: str):
        def local_user_and_host():
            username = 'user'
            hostname = 'host'
            try:
                username = os.environ['USER'] if 'USER' in os.environ else username
                hostname = os.environ['HOSTNAME'] if 'HOSTNAME' in os.environ else hostname
                username = subprocess.check_output(['whoami']).decode('utf8')
                hostname = subprocess.check_output(['hostname']).decode('utf8')
            except Exception as ex:
                log.warning("Failed to obtain user and hostname: {ex}")
            return username, hostname

        totp = pyotp.TOTP(auth_secret)

        # provision URI and QR code
        username, hostname = local_user_and_host()
        uri = totp.provisioning_uri(name=f"{username}@{hostname}", issuer_name=self.params.bot_name)
        sys.stdout.write(f"\n{uri}\n")
        self._dump_qr_code(uri)
        sys.stdout.write(f"TOTP setup key: {totp.secret}\n")
        sys.stdout.write("\n")
        sys.stdout.flush()

    @staticmethod
    def _dump_qr_code(uri: str):
        try:
            import qrcode
            qr = qrcode.QRCode()
            qr.add_data(uri)
            qr.print_tty()
        except ModuleNotFoundError as ex:
            log.warning("qrcode not found, try: `pip install qrcode`")

    def _do_start(self):
        params = TeleRemBot.Params.from_yaml_file(self.params.config)
        bot = TeleRemBot.from_params(params)
        bot.main()

    def _read_artifact(self, name):
        try:
            import importlib.resources as pkg_resources
        except ImportError:
            # for PY<37 try backported `importlib_resources`.
            import importlib_resources as pkg_resources
        from . import artifacts
        content = pkg_resources.read_text(artifacts, name)
        return content

    def _template_substitute(self, name: str, mapping: dict):
        return string.Template(self._read_artifact(name)).safe_substitute(mapping)

    def _escape_bash(self, content: str):
        return re.sub(r'\$(?!\{)', r'\$', content)

    def _do_install(self):
        config_path = self.params.resolve_path(self.params.config)
        if not os.path.exists(config_path):
            log.error(f"config not found, call init first: {config_path}")
            exit(2)
            return
        # dump the install script to stdout
        if not (self.params.systemd ^ self.params.initd):
            log.error("install needs a single --systemd or --initd CLI option")
            exit(2)
            return

        config = BotDaemon.Params.from_dict(self._load_config(), return_unused=False, return_instance=True)
        config.scripts_root = "${TELEREM_SCRIPTS_ROOT}"
        print(self._template_substitute("service-prepare.bash.template", {
            "_TELEREM_USER": config.service_user,
            "_TELEREM_HOME": f"/home/{config.service_user}",
            "_SERVICE_NAME": config.service_name,
            "_WELCOME_SCRIPT_CONTENT": self._escape_bash(self._read_artifact("welcome.sh.template")),
            "_TELEREMD_CONF_YAML_CONTENT": config.to_yaml_string()
        }))
        if self.params.initd:
            print(self._template_substitute("service-initd.bash.template", {
                "_INITD_SCRIPT_CONTENT": self._escape_bash(self._read_artifact("initd.template"))
            }))
        elif self.params.systemd:
            print(self._template_substitute("service-systemd.bash.template", {
                "_SYSTEMD_SERVICE_CONTENT": self._escape_bash(self._read_artifact("systemd.service.template"))
            }))


def main():
    def parse_cli_params(cls):
        args, _ = cls.to_argument_parser().parse_known_args()
        params = cls(args._get_kwargs())
        return params

    cli_cmd = parse_cli_params(BotCli.Params).command
    if cli_cmd in ["init", "start", "install"]:
        params = parse_cli_params(BotCli.Params)
        cli = BotCli.from_params(params)
        cli.main()
    elif cli_cmd in ["daemon"]:
        from telerembash.daemon import BotDaemonCli
        params = parse_cli_params(BotDaemonCli.Params)
        daemon = BotDaemonCli.from_params(params)
        daemon.main()
    else:
        BotCli.Params.to_argument_parser().print_usage()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

#
# created by kpe on 09.01.2021 at 2:20 PM
#

import pathlib
import sys

from setuptools import setup, find_packages, convert_path


WORK_DIR = pathlib.Path(__file__).parent
MOD_NAME = "telerembash"


# Check python version
MINIMAL_PY_VERSION = (3, 7)
if sys.version_info < MINIMAL_PY_VERSION:
    raise RuntimeError(f"{MOD_NAME} works only with Python {{}}+".format('.'.join(map(str, MINIMAL_PY_VERSION))))


def _version():
    ns = {}
    with open(convert_path(f"{MOD_NAME}/version.py"), "r") as fh:
        exec(fh.read(), ns)
    return ns['__version__']


__version__ = _version()


with open("README.rst", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("requirements.txt", "r") as reader:
    install_requires = list(map(lambda x: x.strip(), reader.readlines()))


setup(name=f"{MOD_NAME}",
      version=__version__,
      url="https://github.com/kpe/telerembash/",
      description="A Telegram Remote Shell Bot with HOTP authentication.",
      long_description=long_description,
      long_description_content_type="text/x-rst",
      keywords="telegram totp hotp authentication bot",
      license="MIT",

      author="kpe",
      author_email="kpe.git@gmailbox.org",
      packages=find_packages(exclude=["tests"]),
      package_data={"": ["*.txt", "*.rst"],
                    "telerembash": ["artifacts/*.yaml", "artifacts/*.template"]},
      entry_points={
          'console_scripts': ['telerem=telerembash.cli:main']
      },

      zip_safe=True,
      install_requires=install_requires,
      python_requires=">=3.5",
      classifiers=[
          "Development Status :: 5 - Production/Stable",
          "License :: OSI Approved :: MIT License",
          "Programming Language :: Python",
          "Programming Language :: Python :: 3",
          "Programming Language :: Python :: 3.7",
          "Programming Language :: Python :: 3.8",
          "Programming Language :: Python :: 3.9",
          "Programming Language :: Python :: Implementation :: CPython",
          "Programming Language :: Python :: Implementation :: PyPy"])

